
import pyglet
from pyglet.gl import *
import cocos
from cocos import tiles
from cocos.director import director

class TileEditorControlLayer(cocos.layer.Layer):
    editor = None
    def on_text(self, s):
        if s != 'e': return False
        if self.editor is not None: return False

        editor = TileEditorLayer(self.manager, self.level,
            'data/level-tiles.xml', 'level-tiles', self.edit_complete)
        self.scene.add(editor)
        return True

    def edit_complete(self, layer):
        self.scene.remove(layer)
        self.editor = None


class TileEditorLayer(tiles.ScrollableLayer):
    def __init__(self, manager, map_layer, filename, id, on_done):
        super(TileEditorLayer, self).__init__()
        self.manager = manager
        self.manager.append(self)
        self.map_layer = map_layer
        tileset = cocos.tiles.load(filename)[id]
        self.tilesets = [
           (filename, ''),
        ]

        self.selector = TileSetWidget(tileset)
        self.highlight = None

        @self.selector.window.event
        def on_close():
            self.manager.remove(self)
            self.manager = None
            self.selector.close()
            on_done(self)
            return True

    def on_key_press(self, key, modifier):
        if modifier == pyglet.window.key.MOD_ACCEL:
            if key == pyglet.window.key.S:
                self.map_layer.save_xml(self.tilesets)
                director.pop()
            elif key == pyglet.window.key.Q:
                director.pop()

    def on_text_motion(self, motion):
        fx, fy = self.manager.fx, self.manager.fy
        if motion == pyglet.window.key.MOTION_UP:
            self.manager.force_focus(fx, fy+50)
        elif motion == pyglet.window.key.MOTION_DOWN:
            self.manager.force_focus(fx, fy-50)
        elif motion == pyglet.window.key.MOTION_LEFT:
            self.manager.force_focus(fx-50, fy)
        elif motion == pyglet.window.key.MOTION_RIGHT:
            self.manager.force_focus(fx+50, fy)
        else:
            return False
        return True

    def on_mouse_press(self, x, y, buttons, modifiers):
        m = self.map_layer

        x, y = m.get_virtual_coordinates(x, y)

        i, j = int(x // m.tw), int(y // m.th)
        x, y = m.tw * i, m.th * j

        if i < 0 or j < 0:
            return False

        # add columns?
        N = len(m.cells)
        M = len(m.cells[0])
        if i >= N:
            for n in range(N - i + 1):
                m.cells.append([tiles.RectCell(N+n, p, m.tw, m.th, {}, None)
                    for p in range(M)])

        # add rows?
        N = len(m.cells[0])
        if j >= N:
            diff = j - N + 1
            for n, column in enumerate(m.cells):
                column.extend([tiles.RectCell(n, N+p, m.tw, m.th, {}, None)
                    for p in range(diff)])

        cell = m.get(x, y)

        # now add the cell
        if buttons & pyglet.window.mouse.LEFT:
            cell.tile = self.selector.tileset[self.selector.current.tile_id]
            self.map_layer.set_dirty()
            self.manager.force_focus(self.manager.fx, self.manager.fy)
        elif buttons & pyglet.window.mouse.RIGHT:
            cell.tile = None
            self.map_layer.set_dirty()
            self.manager.force_focus(self.manager.fx, self.manager.fy)
        return True

    def on_mouse_motion(self, x, y, dx, dy):
        x, y = self.map_layer.get_virtual_coordinates(x, y)
        i, j = int(x // self.map_layer.tw), int(y // self.map_layer.th)
        if i < 0 or j < 0:
            self.highlight = None
            return True
        x, y = self.map_layer.tw * i, self.map_layer.th * j
        self.highlight = (x, y, x+self.map_layer.tw, y+self.map_layer.th)
        return True

    def draw(self):
        if self.highlight is None:
            return
        glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glColor4f(1, 1, 0, .3)
        glRectf(*self.highlight)
        glPopAttrib()

class TileSetWidget(object):
    def __init__(self, tileset):
        self.window = pyglet.window.Window(width=256,height=512,
            style=pyglet.window.Window.WINDOW_STYLE_TOOL)
        self.window.push_handlers(self)
        self.tileset = tileset
        x = y = 0
        self.batch = pyglet.graphics.Batch()
        self.sprites = []
        for n, k in enumerate(tileset):
            s = pyglet.sprite.Sprite(tileset[k].image, y=y, x=x,
                batch=self.batch)
            if not n:
                self.current = s
                s.color = (255, 200, 200)
            s.tile_id = k
            self.sprites.append(s)
            s.scale = 64. / s.width
            y += 64
            if y >= 512: y = 0; x += 64

    def on_mouse_press(self, x, y, buttons, modifiers):
        for s in self.sprites:
            if x < s.x or x > s.x + s.width: continue
            if y < s.y or y > s.y + s.height: continue
            self.current.color = (255, 255, 255)
            self.current = s
            s.color = (255, 200, 200)
            return True

    def on_draw(self):
        self.batch.draw()

    def close(self):
        self.window.pop_handlers()
        self.window.close()

